//Generated by lombok.ast.template.TemplateProcessor. DO NOT EDIT, DO NOT CHECK IN!

package lombok.ast;

public class Comment extends lombok.ast.AbstractNode {
	private boolean blockComment = false;
	private java.lang.String content = "";
	
	public boolean astBlockComment() {
		return this.blockComment;
	}
	
	public lombok.ast.Comment astBlockComment(boolean blockComment) {
		this.blockComment = blockComment;
		return this;
	}
	
	public java.lang.String astContent() {
		return this.content;
	}
	
	public lombok.ast.Comment astContent(java.lang.String content) {
		this.content = content== null ? "" : content;
		return this;
	}
	
	@java.lang.Override public java.util.List<Node> getChildren() {
		java.util.List<Node> result = new java.util.ArrayList<Node>();
		return result;
	}
	
	@java.lang.Override public boolean replaceChild(Node original, Node replacement) throws lombok.ast.AstException {
		return false;
	}
	
	@java.lang.Override public boolean detach(Node child) {
		return false;
	}
	
	@java.lang.Override public void accept(lombok.ast.AstVisitor visitor) {
		if (visitor.visitComment(this)) return;
		visitor.afterVisitComment(this);
		visitor.endVisit(this);
	}
	
	@java.lang.Override public Comment copy() {
		Comment result = new Comment();
		result.blockComment = this.blockComment;
		result.content = this.content;
		return result;
	}
	
	public boolean isJavadoc() {
		return lombok.ast.CommentTemplate.isJavadoc(this);
	}
	
	public boolean isMarkedDeprecated() {
		return lombok.ast.CommentTemplate.isMarkedDeprecated(this);
	}
	
}
