class C010_CombinedBinaryExpressions {
    {
        int five = 5;
        int ten = 10;
        int binaryOrCombined = 20 | five | ten;
        int binaryPlusCombined = 20 + five + ten;
        int veryLongCombined = five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten;
        int veryVeryLongCombined = five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten + five + ten;
    }
}