class D001_VariableDeclarations {
    {
        boolean c;
        boolean[] c1;
        boolean c2[];
        String[][] allNames = {{"Reinier", "Roel"}};
        String x, y;
        final int a = 10, b[] = {20};
        int m, n[], o;
        boolean s[], t[];
    }
}