class E006_TypeParametersOnGenericTypesMethodDeclaration<C, X extends Exception> {
    <T extends C, U extends T, E extends Exception> U genericFun(java.util.Map<? extends C, T> arg0) throws E, X {
        return null;
    }
}