class E003_ThrowsConstructorDeclarations1 {
    public E003_ThrowsConstructorDeclarations1() throws Throwable {
    }
}

class E003_ThrowsConstructorDeclarations2 {
    E003_ThrowsConstructorDeclarations2() throws Throwable {
    }
}

class E003_ThrowsConstructorDeclarations3 {
    public E003_ThrowsConstructorDeclarations3() throws NullPointerException, NumberFormatException {
    }
}

class E003_ThrowsConstructorDeclarations4 {
    E003_ThrowsConstructorDeclarations4() throws NullPointerException, NumberFormatException {
    }
}