class E001_ModifiersConstructorDeclarations1 {
    public E001_ModifiersConstructorDeclarations1() {
    }
}

class E001_ModifiersConstructorDeclarations2 {
    protected E001_ModifiersConstructorDeclarations2() {
    }
}

class E001_ModifiersConstructorDeclarations3 {
    E001_ModifiersConstructorDeclarations3() {
    }
}

class E001_ModifiersConstructorDeclarations4 {
    private E001_ModifiersConstructorDeclarations4() {
    }
}