class G003_MultipleFieldDeclaration {
    private int x = 0, y;
    
    int z, k[] = {};
    
    boolean[] a[][], b[];
    
    @SuppressWarnings("all")
    public final String[] name = null, names[] = {{}};
}