//Generated by lombok.ast.template.TemplateProcessor. DO NOT EDIT, DO NOT CHECK IN!

package lombok.ast;

public class Identifier extends lombok.ast.AbstractNode implements lombok.ast.DescribedNode {
	private java.lang.String value = "";
	
	public java.lang.String astValue() {
		return this.value;
	}
	
	public lombok.ast.Identifier astValue(java.lang.String value) {
		this.value = value== null ? "" : value;
		return this;
	}
	
	@java.lang.Override public java.util.List<Node> getChildren() {
		java.util.List<Node> result = new java.util.ArrayList<Node>();
		return result;
	}
	
	@java.lang.Override public boolean replaceChild(Node original, Node replacement) throws lombok.ast.AstException {
		return false;
	}
	
	@java.lang.Override public boolean detach(Node child) {
		return false;
	}
	
	@java.lang.Override public void accept(lombok.ast.AstVisitor visitor) {
		if (visitor.visitIdentifier(this)) return;
		visitor.afterVisitIdentifier(this);
		visitor.endVisit(this);
	}
	
	@java.lang.Override public Identifier copy() {
		Identifier result = new Identifier();
		result.value = this.value;
		return result;
	}
	
	public java.lang.String getDescription() {
		return lombok.ast.IdentifierTemplate.getDescription(this);
	}
	
	public static Identifier of(java.lang.String value) {
		return lombok.ast.IdentifierTemplate.of(value);
	}
	
}
