class I001_Deprecated {
    @Deprecated
    int x;
    
    /**
     * @deprecated
     */
    int y;
    
    /**
     * @deprecated
     */
    class Inner {
        /**
         * @deprecated
         */
        Inner() {
        }
        
        /**
         * @deprecated
         */
        int test() {
            return 0;
        }
    }
}