class C002_TextLiterals {
    {
        char a = 'a';
        char b = '\n';
        char c = '\t';
        char d = '\0';
        char e = '\10';
        char f = '\\';
        char g = '\'';
        String z = "";
        String y = "Simple string";
        String w = "With escapes \\ \" \n \t \0 \7 \10 \11 \f \r ";
        String v = ";";
    }
}