class D001_BasicStatements {
    {
        int a = 10;
        {
            a++;
            --a;
            "a".toString();
            a = 20;
            ++a;
            a--;
            new Object();
        }
        ;
        ;
        {
        }
    }
}