class B004_1<T> {
}

class B004_2<T extends java.lang.Number> {
}

class B004_3<A, B> {
}

class B004_4<A extends java.util.List<String> & java.util.RandomAccess> {
}

class B004_5<A extends java.util.List<?>> {
}

class B004_6<T, A extends java.util.List<? extends T>> {
}