/**
 * jdoc1
 * multiliner
 */
class H010_Javadoc {
    /** jdoc2 */
    int x;
    
    /** jdoc3foo */
    class A {
    }
    
    /** jdoc4 */
    interface B {
    }
    
    /** jdoc5 */
    @interface C {
        /** jdoc8 */
        int foo() default 5;
    }
    
    /** jdoc6 */
    enum D {
    }
    
    /** jdoc7
      * @throws NumberFormatException foo
     */
    void test() {
    }
    
    /** {@inheritDoc} */
    void test2() {
    }
}