class B005_ArrayTypes {
    {
        int[] simple;
        int postfix[];
        int[] postfix2[];
        java.lang.String[] oneDimensional;
        java.lang.String[][] twoDimensional;
        java.lang.String twoDimensional2[][];
        java.util.List<int[]> generified;
        java.util.List<String>[] generifiedArraysAreABadIdeaButLegal;
    }
}