/*
 * Copyright (C) 2010 The Project Lombok Authors.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package lombok.ast.syntaxChecks;

import static lombok.ast.syntaxChecks.MessageKey.*;
import static lombok.ast.Message.*;

import lombok.ast.Block;
import lombok.ast.Break;
import lombok.ast.Continue;
import lombok.ast.Node;
import lombok.ast.Return;
import lombok.ast.Statement;
import lombok.ast.template.SyntaxCheck;

@SyntaxCheck
public class UnreachableStatementsChecks {
	public void unreachablesAfterBreak(Break statement) {
		checkForUnreachables(statement);
	}
	
	public void unreachablesAfterContinue(Continue statement) {
		checkForUnreachables(statement);
	}
	
	public void unreachableAfterReturn(Return statement) {
		checkForUnreachables(statement);
	}
	
	private void checkForUnreachables(Statement n) {
		Block b = n.upToBlock();
		if (b == null) return;
		
		boolean found = false;
		for (Node s : b.rawContents()) {
			if (found) {
				s.addMessage(error(STATEMENT_UNREACHABLE, "Unreachable code"));
				return;
			}
			if (s == n) found = true;
		}
	}
}
