class C001_NumberLiterals {
    {
        int x = 0;
        int y = 1000000;
        long z = 100000000000L;
        int a = 0x1fa1;
        int b = 0100;
        double c = 100.0;
        double d = 0.0;
        double e = 123e10;
        double f = 123e-10;
        double g = 123e+10;
        double h = 0x1p1;
        float i = 123F;
        float j = 0xap-1F;
        boolean b1 = true;
        boolean b2 = false;
        Object o = null;
    }
}